"
 COPYRIGHT (c) 1988-93 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Object subclass:#OperatingSystem
       instanceVariableNames:''
       classVariableNames:''
       poolDictionaries:''
       category:'System-Support'
!

OperatingSystem comment:'

COPYRIGHT (c) 1988-93 by Claus Gittinger
             All Rights Reserved

this class gives access to some operating system services;
some of it is very specific for unix. 

%W% %E%

written 1988 by claus
'!

%{

#ifdef transputer
# define unlink(f)      ((remove(f) == 0) ? 0 : -1)
#else
# include <signal.h>

# ifdef SYSV
#  include <sys/types.h>
#  include <sys/param.h>
#  include <sys/times.h>
#  if ! defined(sco3_2)
#   include <unistd.h>
#  endif
#  if defined(isc3_2) || defined(sco3_2)
#   include <sys/time.h>
#  endif
#  if !defined(isc3_2)
#   if defined(PCS) && defined(mips)
#    include "/usr/include/bsd/sys/time.h"
#    include "/usr/include/sys/time.h"
#   else
#    include <time.h>
#   endif
#  endif
#  if defined(isc3_2)
#   include <sys/bsdtypes.h>
#  endif
#  ifdef FAST_TIMER
#   include <ft.h>
    static int timer_fd = -1;
#  endif
# else /* not SYSV */
#  include <sys/time.h>
#  include <sys/types.h>
# endif
# include <pwd.h>
# include <grp.h>

# include <sys/stat.h>
# ifndef S_IXUSR
#  define S_IXUSR S_IEXEC
#  define S_IXGRP (S_IEXEC>>3)
#  define S_IXOTH (S_IEXEC>>6)
# endif

#endif

#ifndef errno
 extern errno;
#endif

#include <stdio.h>
#ifndef R_OK
# define R_OK    4       /* Test for Read permission */ 
# define W_OK    2       /* Test for Write permission */
# define X_OK    1       /* Test for eXecute permission */
# define F_OK    0       /* Test for existence of File */
#endif
%}

!OperatingSystem class methodsFor:'misc'!

exit
    "shutdown smalltalk immediately"

%{  /* NOCONTEXT */
    mainExit(0);
%}
    "OperatingSystem exit - dont evaluate this"
!

exit:exitCode
    "shutdown smalltalk immediately returning an exit-code"

%{  /* NOCONTEXT */
    if (! _isSmallInteger(exitCode))
        exit(1);
    mainExit(_intVal(exitCode));
%}
    "OperatingSystem exit:1 - dont evaluate this"
!

getEnvironment:aString
    "get an environment string"

%{  /* NOCONTEXT */

    char *env;

    if (_isString(aString)) {
        env =  (char *)getenv(_stringVal(aString));
        if (env) {
            RETURN ( _MKSTRING(env COMMA_CON) );
        }
    }
%}
.
    ^ nil

    "OperatingSystem getEnvironment:'LANG'"
!

getLoginName
    "return a string with the users name"

%{  /* NOCONTEXT */

    char *name = "you";
#ifndef transputer
    name = (char *)getlogin();
    if (! name) {
        name = (char *)getenv("LOGNAME");
    }
#endif
    RETURN ( _MKSTRING(name COMMA_CON) );
%}
    "OperatingSystem getLogin"
!

getUserNameFromID:aNumber
    "return the user-name-string for a given numeric user-id"

%{  /* NOCONTEXT */

#ifndef transputer
    struct passwd *p;

    if (_isSmallInteger(aNumber)) {
        p = getpwuid(_intVal(aNumber));
        if (p) {
            RETURN ( _MKSTRING(p->pw_name COMMA_CON) );
        }
    }
#endif
%}
.
    ^ '???'

    "OperatingSystem getUserNameFromID:0"
!

getGroupNameFromID:aNumber
    "return the group-name-string for a given numeric group-id"

%{  /* NOCONTEXT */

#ifndef transputer
    struct group *g;

    if (_isSmallInteger(aNumber)) {
        g = getgrgid(_intVal(aNumber));
        if (g) {
            RETURN ( _MKSTRING(g->gr_name COMMA_CON) );
        }
    }
#endif
%}
.
    ^ '???'

    "OperatingSystem getGroupNameFromID:0"
!

getHomeDirectory
    "return the name of the home directory"

    ^ OperatingSystem getEnvironment:'HOME'

    "OperatingSystem getHomeDirectory"
!

getProcessId
    "return the processId"

%{  /* NOCONTEXT */

    int pid = 0;
#ifndef transputer
    pid = getpid();
#endif
    RETURN ( _MKSMALLINT(pid) );
%}
    "OperatingSystem getProcessId"
!

getCharacter
    "read a character from keyboard"

%{  /* NOCONTEXT */

    RETURN ( _MKSMALLINT(getchar()) );
%}
!

getCPUType
    "return a string giving the type of machine we're running on"

    |cpu|
    
    cpu := 'unknown'.

%{  /* NOCONTEXT */
#ifdef vax
    cpu = _MKSTRING("vax" COMMA_CON);
#endif
#ifdef mips
    cpu = _MKSTRING("mips" COMMA_CON);
#endif
#ifdef i386
    cpu = _MKSTRING("i386" COMMA_CON);
#endif
#ifdef ns32k
    cpu = _MKSTRING("ns32k" COMMA_CON);
#endif
#ifdef mc68k
    cpu = _MKSTRING("mc68k" COMMA_CON);
#endif
#ifdef sparc
    cpu = _MKSTRING("sparc" COMMA_CON);
#endif
#ifdef snake
    cpu = _MKSTRING("snake" COMMA_CON);
#endif
#ifdef rs6000
    cpu = _MKSTRING("rs6000" COMMA_CON);
#endif
#ifdef alpha
    cpu = _MKSTRING("alpha" COMMA_CON);
#endif
#ifdef transputer
    cpu = _MKSTRING("transputer" COMMA_CON);
#endif
%}
.
    ^ cpu

    "OperatingSystem getCPUType"
!

getOSType
    "return a string giving the type of os we're running on"

    |os|

    os := 'unknown'.

%{  /* NOCONTEXT */
#ifdef sinix
  os = _MKSTRING("sinix" COMMA_CON);
#endif

#ifdef ultrix
  os = _MKSTRING("ultrix" COMMA_CON);
#endif

#ifdef sco
  os = _MKSTRING("sco" COMMA_CON);
#endif

#ifdef LINUX
  os = _MKSTRING("linux" COMMA_CON);
#endif

#ifdef BSD
# ifdef MACH
  os = _MKSTRING("mach" COMMA_CON);
# endif

# ifdef sunos
  os = _MKSTRING("sunos" COMMA_CON);
# endif

# ifdef IRIS
  os = _MKSTRING("irix" COMMA_CON);
# endif

  if (os == nil) os = _MKSTRING("bsd" COMMA_CON);
#endif

#ifdef SYSV
# ifdef SYSV3
  os = _MKSTRING("sys5.3" COMMA_CON);
# else
#  ifdef SYSV4
    os = _MKSTRING("sys5.4" COMMA_CON);
#  else
    os = _MKSTRING("sys5" COMMA_CON);
#  endif
# endif
#endif
%}
.
    ^ os

    "OperatingSystem getOSType"
!

getSystemType
    "return a string giving the type of system we're running on"

    |sys|

    sys := 'unknown'.

%{  /* NOCONTEXT */
#ifdef sinix
  sys = _MKSTRING("sinix" COMMA_CON);
#endif

#ifdef ultrix
  sys = _MKSTRING("ultrix" COMMA_CON);
#endif

#ifdef sco
  sys = _MKSTRING("sco" COMMA_CON);
#endif

#ifdef sunos
  sys = _MKSTRING("sunos" COMMA_CON);
#endif

#ifdef solaris
  sys = _MKSTRING("solaris" COMMA_CON);
#endif

#ifdef NEXT
  sys = _MKSTRING("next" COMMA_CON);
#endif

#ifdef IRIS
  sys = _MKSTRING("iris" COMMA_CON);
#endif

#ifdef LINUX
  sys = _MKSTRING("linux" COMMA_CON);
#endif

#ifdef BSD
# ifdef MACH
  if (sys == nil) sys = _MKSTRING("mach" COMMA_CON);
# endif
  if (sys == nil) sys = _MKSTRING("bsd" COMMA_CON);
#endif

#ifdef SYSV
# ifdef SYSV3
  if (sys == nil) sys = _MKSTRING("sys5.3" COMMA_CON);
# else
#  ifdef SYSV4
    if (sys == nil) sys = _MKSTRING("sys5.4" COMMA_CON);
#  else
    if (sys == nil) sys = _MKSTRING("sys5" COMMA_CON);
#  endif
# endif
#endif
%}
.
    ^ sys

    "OperatingSystem getSystemType"
!

isBSDlike
    "return true, if the OS we're running on is a real unix."

%{  /* NOCONTEXT */

#ifdef BSD
    RETURN ( true );
#endif
#ifdef SYSV4
    RETURN ( true );
#endif
%}
.
    ^ false
!

maxFileNameLength
    "return the max number of characters in a filename."

%{  /* NOCONTEXT */
#if defined(BSD) || defined(SYSV4) || defined(LINUX)
    RETURN ( _MKSMALLINT(255) );
#else
# ifdef SYSV
    RETURN ( _MKSMALLINT(14) );
# endif
# ifdef MSDOS
    RETURN ( _MKSMALLINT(9) );
# endif
#endif
%}
.
    ^ 14
! !

!OperatingSystem class methodsFor:'error messages'!

errorTextForNumber:errNr
    "return a message string from a unix errorNumber 
     (as returned by a system call). Should be replaced by
     a resource lookup."

    |msg messages|

    (Language == #german) ifTrue:[
        messages := #('keine superuser Berechtigung'
                      'ungueltiger Datei- oder VerzeichnisName'
                      nil "'ungueltige Prozessnummer'  "
                      nil "'unterbrochener systemcall' "
                      'E/A Fehler'
                      nil "'Geraet existiert nicht' "
                      'zu viele Argumente'
                      'nicht ausfuehrbar'
                      nil "'falscher FileDescriptor'"
                      nil "'kein Kindprozess'       "
                      'zu viele Prozesse oder zu wenig Speicher'
                      'zu wenig Speicher'
                      'keine ZugriffsBerechtigung'
                      nil "'falsche Adresse'        "
                      nil "'kein Blockgeraet'       "
                      nil "'Platte noch im Zugriff' "
                      'Datei existiert bereits'
                      nil "'Link ueber Plattengrenzen hinweg' "
                      'Geraet existiert nicht'
                      'ist kein Verzeichnis'
                      'ist ein Verzeichnis'
                      nil "'ungueltiges Argument' "
                      'zu viele Dateien offen'
                      'zu viele Dateien offen'
                      nil "'kein Terminalgeraet'  "
                      'Datei wird gerade ausgefuehrt'
                      'Datei zu gross'
                      'Platte ist voll'
                      'ungueltige Positionierung'
                      'Platte ist schreibgeschuetzt'
                      'zu viele Links'
                      'Pipe unterbrochen'
                      'argument nicht im gueltigen Bereich'
                      'Ergebnis nicht im gueltigen Bereich')
    ] ifFalse:[
        messages := #('Not super-user'
                      'No such file or directory'
                      nil "'No such process'   "
                      nil "'interrupted system call' "
                      'I/O error'
                      nil "'No such device or address' "
                      'Arg list too long'
                      'Exec format error'
                      nil "'Bad file number'"
                      nil "'No children'       "
                      'No more processes'
                      'Not enough core'
                      'Permission denied'
                      nil "'Bad address'        "
                      nil "'Block device required'       "
                      nil "'Mount device busy' "
                      'File exists'
                      nil "'Cross-device link' "
                      'No such device'
                      'Not a directory'
                      'Is a directory'
                      nil 'Invalid argument'
                      'File table overflow'
                      'Too many open files'
                      nil "'Not a typewriter' "
                      'Text file busy'
                      'File too large'
                      'No space left on device'
                      'Illegal seek'
                      'Read only file system'
                      'Too many links'
                      'Broken pipe'
                      'Math arg out of domain of func'
                      'Math result not representable')
    ].

    (errNr between:1 and:messages size) ifTrue:[
        msg := messages at:errNr
    ].
    msg isNil ifTrue:[
        ^ ('ErrorNr: ' , errNr printString)
    ].
    ^ msg
! !

!OperatingSystem class methodsFor:'interrupts'!

enableUserInterrupts
    "enable userInterrupt (^C) handling;
     after enabling, ^C will send the message 'userInterrupt'
     to the UserInterruptHandler object."

%{  /* NOCONTEXT */
    extern void userInterrupt(), exceptionInterrupt();

    signal(SIGINT, userInterrupt);
    /* signal(SIGQUIT, userInterrupt); */
%}
!

enableFpExceptionInterrupts
    "enable floating point exception interrupts (if the
     architecture supports it).
     after enabling, fpu-exceptions will send the message 
     'fpuExceptionInterrupt' to the FPUExceptionInterruptHandler object."

%{  /* NOCONTEXT */
    extern void fpExceptionInterrupt();

    signal(SIGFPE, fpExceptionInterrupt);
%}
!

enableSignalInterrupts
    "enable signal exception interrupts (trap, buserror & segm. violation).
     after enabling, these exceptions will send the message 
     'signalInterrupt' to the SignalInterruptHandler object."

%{  /* NOCONTEXT */
    extern void signalPIPEInterrupt();
    extern void signalBUSInterrupt();
    extern void signalSEGVInterrupt();

    signal(SIGPIPE, signalPIPEInterrupt);
#ifdef SIGBUS
    signal(SIGBUS,  signalBUSInterrupt);
#endif
    signal(SIGSEGV, signalSEGVInterrupt);
%}
!

enableIOInterrupts
    "enable IO availability interrupts 
     (SIGPOLL/SIGIO, if the architecture supports it).
     after enabling, these signals will send the message 
     'ioInterrupt' to the IOInterruptHandler object."

%{  /* NOCONTEXT */
    extern void ioInterrupt();

#ifdef SIGPOLL
    signal(SIGPOLL, ioInterrupt);
#endif
#ifdef SIGIO
    signal(SIGIO,  ioInterrupt);
#endif
%}
!

startSpyTimer
    "trigger a spyInterrupt, to be signalled after some (short) time.
     This is used by MessageTally for profiling."

%{  /* NOCONTEXT */

    extern void spyInterrupt();
#ifdef FAST_TIMER
    /*
     * using PD ft-driver for tick-resolution signals
     */
    if (timer_fd < 0) {
        timer_fd = open("/dev/ft0", 0);
    }
    if (timer_fd > 0) {
        ioctl(timer_fd, FTIOCSET, (2<<16) | SIGALRM);
    }
#else
# if defined(BSD) || defined(isc3_2) || defined(SYSV4) || defined(LINUX)
    struct itimerval dt;

    dt.it_interval.tv_sec = 0;
    dt.it_interval.tv_usec = 0;
    dt.it_value.tv_sec = 0;
    dt.it_value.tv_usec = 10000;   /* 100 Hz */
    setitimer(ITIMER_VIRTUAL, &dt, 0);
#  ifdef BSD
#   ifndef SYSV4
    sigsetmask(0);
#   endif
#  endif
# endif
#endif

#ifdef SIGVTALRM
    signal(SIGVTALRM, spyInterrupt);
#else
    signal(SIGALRM, spyInterrupt);
#endif
%}
!

stopSpyTimer
    "stop spy timing"

%{  /* NOCONTEXT */

#ifdef FAST_TIMER
    if (timer_fd > 0) {
        ioctl(timer_fd, FTIOCANCEL, 0);
        close(timer_fd);
        timer_fd = -1;
    }
#else
# if defined(BSD) || defined(isc3_2) || defined(SYSV4) || defined(LINUX)
    struct itimerval dt;

    dt.it_interval.tv_sec = 0;
    dt.it_interval.tv_usec = 0;
    dt.it_value.tv_sec = 0;
    dt.it_value.tv_usec = 0;
    setitimer(ITIMER_VIRTUAL, &dt, 0);
# endif
#endif
%}
! !

!OperatingSystem class methodsFor:'time and date'!

getTimeLow
    "return low 16 bits of current time. 
     Obsolete: Dont use this method, use getTimeParts below.
     This method will not always return the correct time 
     if used together with getTimeHi.
     (a wrap between the two getTimeXXX calls could occur)"

%{  /* NOCONTEXT */

    RETURN ( _MKSMALLINT(time(0) & 0xFFFF) );
%}
.
    self primitiveFailed

    "OperatingSystem getTimeLow"
!

getTimeHi
    "return hi 16 bits of current time. 
     Obsolete: Dont use this method, use getTimeParts below. 
     This method will NOT always return the correct time
     if used together with getTimeHi.
     (a wrap between the two getTimeXXX calls could occur)"

%{  /* NOCONTEXT */

    RETURN ( _MKSMALLINT((time(0) >> 16) & 0xFFFF) );
%}
.
    self primitiveFailed

    "OperatingSystem getTimeHi"
!

getTimeInto:aBlock
    "evaluate the argument aBlock, passing the time-parts of
     the current time as arguments."

    |low hi|
%{ 
    int now;

    now = time(0);
    hi  = _MKSMALLINT((now >> 16) & 0xFFFF);
    low = _MKSMALLINT(now & 0xFFFF);
%}
.
    aBlock value:low value:hi

    "OperatingSystem getTimeTimeInto:[:low :hi | low printNewline. hi printNewline]"
!

getTime
    "return current Time (in seconds since 1970).
     This might return a LargeInteger some time."

    ^ self getTimeHi * 16r10000 + self getTimeLow

    "OperatingSystem getTime"
!

computeDatePartsOf:timeLow and:timeHi for:aBlock
    "compute year, month and day from the time-parts timeLow and
     timeHi and evaluate the argument, a 3-arg block with these.
     This method was added to avoid LargeInteger arithmetic; the time-parts
     are those returned by getTimeLow and getTimeHi."

    |year month day|

    ((timeLow isMemberOf:SmallInteger) and:[timeHi isMemberOf:SmallInteger])
    ifFalse:[
        ^ self primitiveFailed
    ].
%{
    struct tm *tmPtr;
    long t;

    t = (_intVal(timeHi) << 16) | _intVal(timeLow);
    tmPtr = localtime(&t);
    year = _MKSMALLINT(tmPtr->tm_year + 1900);
    month = _MKSMALLINT(tmPtr->tm_mon + 1);
    day = _MKSMALLINT(tmPtr->tm_mday);
%}
.
    aBlock value:year value:month value:day
!

computeTimePartsOf:timeLow and:timeHi for:aBlock
    "compute hours, minutes and seconds from the time-parts timeLow and
     timeHi and evaluate the argument, a 3-arg block with these."

    |hours minutes seconds|

    ((timeLow isMemberOf:SmallInteger) and:[timeHi isMemberOf:SmallInteger])
    ifFalse:[
        ^ self primitiveFailed
    ].
%{
    struct tm *tmPtr;
    long t;

    t = (_intVal(timeHi) << 16) | _intVal(timeLow);
    tmPtr = localtime(&t);
    hours = _MKSMALLINT(tmPtr->tm_hour);
    minutes = _MKSMALLINT(tmPtr->tm_min);
    seconds = _MKSMALLINT(tmPtr->tm_sec);
%}
.
    aBlock value:hours value:minutes value:seconds
!

getMillisecondTime
    "since range is limited to 0..1ffffff and value is wrapping around
     at 1fffffff, this can only be used for relative time deltas.
     Use methods below to compare and add time deltas (should move to Time)"

%{  /* NOCONTEXT */

    long t;
#ifdef SYSV
# ifdef HZ
    /* sys5 time */
    long ticks;
    struct tms tb;

    ticks = times(&tb);
    t = (ticks * 1000) / HZ;
# endif
#else
    /* bsd time */
    struct timeval tb;
    struct timezone tzb;

    gettimeofday(&tb, &tzb);
    t = tb.tv_sec*1000 + tb.tv_usec/1000;
#endif
    RETURN ( _MKSMALLINT(t & 0x0FFFFFFF) );
%}
.
    self error:'time not available'
!

millisecondTimeDeltaBetween:msTime1 and:msTime2
    "subtract two millisecond times (such as returned getMillisecondTime).
     The returned value is msTime1 - msTime2 where a wrap occurs at:16r0FFFFFFF."

    (msTime1 > msTime2) ifTrue:[
        ^ msTime1 - msTime2
    ].
    ^ msTime1 + 16r10000000 - msTime2
!

millisecondTime:msTime1 isAfter:msTime2
    "return true if msTime1 is after msTime2, false if not.
     handling wrap at 16r0FFFFFFF. The two arguments are
     millisecond times (such as returned getMillisecondTime)."

    (msTime1 > msTime2) ifTrue:[
        ((msTime1 - msTime2) > 16r08000000) ifTrue:[
            ^ false
        ].
        ^ true
    ].
    ((msTime2 - msTime1) > 16r08000000) ifTrue:[
        ^ true
    ].
    ^ false
!

millisecondTimeAdd:msTime1 and:msTime2
    "add two millisecond times (such as returned getMillisecondTime).
     The returned value is msTime1 + msTime2 where a wrap occurs at:16r0FFFFFFF."

    |sum|

    sum := msTime1 + msTime2.
    (sum > 16r0FFFFFFF) ifTrue:[^ sum - 16r10000000].
    (sum < 0) ifTrue:[^ sum + 16r10000000].
    ^ sum
!

millisecondDelay:millis
    "delay execution for millis milliseconds."

    self selectOn:nil withTimeOut:(millis * 0.001)

    "OperatingSystem millisecondDelay:1000"
!

sleep:numberOfSeconds
    "cease any action for some time. 
     Not really useful since not even low-prio processes and interrupt
     handling will run during the sleep - use millisecondDelay:."

%{  /* NOCONTEXT */

    if (_isSmallInteger(numberOfSeconds)) {
        sleep(_intVal(numberOfSeconds));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

selectOn:aFileDescriptor withTimeOut:seconds
    "wait for aFileDesriptor to become ready; timeout after t seconds.
     Return true, if i/o ok, false if timed-out or interrupted.
     With 0 as timeout argument, this can be used to check for availability
     of read-data.
     Experimental."

    |millis|

    millis := (seconds * 1000) rounded.
%{
    fd_set rset, wset, eset;
    int t, fd, i, lX, bX;
    struct timeval wt;

    if ((aFileDescriptor == nil) || _isSmallInteger(aFileDescriptor)) {
        if (_isSmallInteger(millis)) {
            FD_ZERO(&rset);
            FD_ZERO(&wset);
            FD_ZERO(&eset);
            if (aFileDescriptor != nil) {
                fd = _intVal(aFileDescriptor);
                if ((fd >= 0) && (fd < FD_SETSIZE))
                    FD_SET(fd, &rset);
            } else
                fd = 0;
            t = _intVal(millis);
            wt.tv_sec = t / 1000;
            wt.tv_usec = (t % 1000) * 1000;
            RETURN ( (select(fd+1, &rset, &wset, &eset, &wt) == 0) ? false
                                                                   : true );
        }
    }
%}
.
    self primitiveFailed
!

selectOn:fd1 and:fd2 withTimeOut:seconds
    "wait for any fd to become ready; timeout after t seconds.
     Return fd if i/o ok, nil if timed-out or interrupted.
     Experimental."

    |millis|

    millis := (seconds * 1000) rounded.
%{
    fd_set rset, wset, eset;
    int t, f1, f2, i, lX, bX;
    struct timeval wt;
    OBJ retFd;

    if (((fd1 == nil) || _isSmallInteger(fd1)) 
     && ((fd2 == nil) || _isSmallInteger(fd2)))  {
        if (_isSmallInteger(millis)) {
            FD_ZERO(&rset);
            FD_ZERO(&wset);
            FD_ZERO(&eset);
            if (fd1 != nil) {
                f1 = _intVal(fd1);
                if ((f1 >= 0) && (f1 < FD_SETSIZE))
                    FD_SET(f1, &rset);
            } else
                f1 = 0;
            if (fd2 != nil) {
                f2 = _intVal(fd2);
                if ((f2 >= 0) && (f2 < FD_SETSIZE))
                    FD_SET(f2, &rset);
            } else
                f2 = 0;
            if (f2 > f1)
                f1 = f2;
            t = _intVal(millis);
            wt.tv_sec = t / 1000;
            wt.tv_usec = (t % 1000) * 1000;
            if (select(f1+1, &rset, &wset, &eset, &wt)) {
                if (FD_ISSET(f1, &rset)) retFd = fd1;
                else if (FD_ISSET(f1, &wset)) retFd = fd1;
                else if (FD_ISSET(f1, &eset)) retFd = fd1;
                else if (FD_ISSET(f2, &rset)) retFd = fd2;
                else if (FD_ISSET(f2, &wset)) retFd = fd2;
                else if (FD_ISSET(f2, &eset)) retFd = fd2;
                RETURN ( retFd );
            }
            RETURN ( nil );
        }
    }
%}
.
    self primitiveFailed
!

selectOnAnyReadable:fdArray withTimeOut:seconds
    "wait for any fd in fdArray (an Array of integers) to become ready;
     timeout after t seconds. An empty set will always wait.
     Return first ready fd if i/o ok, nil if timed-out or interrupted.
     Experimental."

    |millis count|

    millis := (seconds * 1000) rounded asInteger.
    (fdArray class == Array) ifFalse:[
        ^ self error:'argument must be an Array'
    ].
    count := fdArray size.
%{
    fd_set rset, wset, eset;
    int t, f, maxF, i, lX, bX;
    struct timeval wt;
    OBJ fd, retFd;

    if (_isSmallInteger(millis)) {
        FD_ZERO(&rset);
        FD_ZERO(&wset);
        FD_ZERO(&eset);

        maxF = 0;
        for (i=0; i<_intVal(count);i++) {
            fd = _ArrayInstPtr(fdArray)->a_element[i];
            if (fd != nil) {
                f = _intVal(fd);
                if ((f >= 0) && (f < FD_SETSIZE)) {
                    FD_SET(f, &rset);
/*                   FD_SET(f, &wset);       */
/*                   FD_SET(f, &eset);       */
                    if (f > maxF) maxF = f;
                }
            }
        }
        t = _intVal(millis);
        wt.tv_sec = t / 1000;
        wt.tv_usec = (t % 1000) * 1000;
        if (select(maxF+1, &rset, &wset, &eset, &wt)) {
            for (i=0; i <= maxF; i++) {
                if (FD_ISSET(i, &rset)
/*                 || FD_ISSET(i, &wset)
                   || FD_ISSET(i, &eset) */ ) {
                    RETURN ( _MKSMALLINT(i) );
                }
            }
        }
        RETURN ( nil );
    }
%}
.
    self primitiveFailed
! !

!OperatingSystem class methodsFor:'executing commands'!

fork
    "fork a new process"

%{  /* NOCONTEXT */

    int pid;

    pid = fork();
    RETURN ( _MKSMALLINT(pid) );
%}
.
    self primitiveFailed
!

exec:aPath withArguments:argArray
    "execute the unix command specified by the argument, aPath, with
     arguments in argArray.
     If successful, this method does not return and smalltalk is gone.
     If not sucessfull, false is returned. Normal use is with fork."

%{
    char *argv[64];
    int nargs, i;
    OBJ arg;

    if (_isString(aPath) && _isArray(argArray)) {
	nargs = _arraySize(argArray);
        for (i=0; i < nargs; i++) {
            arg = _ArrayInstPtr(argArray)->a_element[i];
	    if (_isString(arg)) {
                argv[i] = (char *) _stringVal(arg);
	    }
        }
        argv[i] = NULL;
        execv(_stringVal(aPath), argv);
        /* should not be reached */
        RETURN ( false );
    }
%}
.
    self primitiveFailed
!

executeCommand:aCommandString
    "execute the unix command specified by the argument, aCommandString.
     Return true if successful, false otherwise. Smalltalk is suspended,
     while the command is executing."

%{  /* NOCONTEXT */

    int status;
    extern OBJ ErrorNumber;

    if (_isString(aCommandString)) {
        status = system((char *) _stringVal(aCommandString));
        if (status == 0) {
            RETURN ( true );
        }
        ErrorNumber = _MKSMALLINT(errno);
        RETURN ( false );
    }
%}
.
    self primitiveFailed

    "OperatingSystem executeCommand:'pwd'"
    "OperatingSystem executeCommand:'ls -l'"
    "OperatingSystem executeCommand:'invalidCommand'"
! !

!OperatingSystem class methodsFor:'file access'!

baseNameOf:aPath
    "return the baseName of the argument, aPath
     - thats the file/directory name without leading parent-dirs
     (i.e. OperatingSystem baseNameOf:'/usr/lib/st/file' -> 'file'
       and OperatingSystem baseNameOf:'/usr/lib' -> lib).
     This method does not check if the path is valid (i.e. if these directories
     really exist)."

    |prev index|

    (aPath = '/') ifTrue:[^ aPath].
    prev := 1.
    [true] whileTrue:[
        index := aPath indexOf:$/
                    startingAt:prev
                      ifAbsent:[^ aPath copyFrom:prev].
        prev := index + 1
    ]
!

directoryNameOf:aPath
    "return the directoryName of the argument, aPath
     - thats the name of the directory where aPath is
     (i.e. OperatingSystem directoryNameOf:'/usr/lib/st/file' -> '/usr/lib/st'
       and OperatingSystem directoryNameOf:'/usr/lib' -> /usr').
     This method does not check if the path is valid (i.e. if these directories
     really exist)."

    |last|

    (aPath = '/') ifTrue:[^ aPath].
    (aPath startsWith:'/') ifFalse:[
        (aPath endsWith:'/') ifTrue:[
            ^ aPath copyFrom:1 to:(aPath size - 1)
        ].
    ].
    last := 1.
    [true] whileTrue:[
        last := aPath indexOf:$/
                   startingAt:(last + 1)
                     ifAbsent:[(last == 1) ifTrue:[^ '/'].
                               ^ aPath copyFrom:1 to:(last - 1)
                              ]
    ]

    "OperatingSystem directoryNameOf:'/fee/foo/bar'"
    "OperatingSystem directoryNameOf:'foo/bar'"
    "OperatingSystem directoryNameOf:'../../foo/bar'"
!

isValidPath:aPathName
    "return true, if 'aPathName' is a valid path name
     (i.e. the file or directory exists)"

%{  /* NOCONTEXT */

    struct stat buf;

    if (_isString(aPathName)) {
        RETURN ( (stat((char *) _stringVal(aPathName), &buf) < 0) ? false : true );
    }
%}
.
    self primitiveFailed
!

isDirectory:aPathName
    "return true, if 'aPathName' is a valid directory path name.
     (i.e. exists and is a directory)"

%{  /* NOCONTEXT */

    struct stat buf;

    if (_isString(aPathName)) {
        if ((stat((char *) _stringVal(aPathName), &buf) < 0)
         || ((buf.st_mode & S_IFMT) != S_IFDIR)) {
            RETURN ( false );
        }
        RETURN ( true );
    }
%}
.
    self primitiveFailed
!

isReadable:aPathName
    "return true, if the file/dir 'aPathName' is readable."

%{  /* NOCONTEXT */

    extern OBJ ErrorNumber;

    if (_isString(aPathName)) {
        if (access(_stringVal(aPathName), R_OK) == 0) {
            RETURN ( true );
        }
        ErrorNumber = _MKSMALLINT(errno);
        RETURN ( false );
    }
%}
.
    self primitiveFailed
!

isWritable:aPathName
    "return true, if the given file is writable"

%{  /* NOCONTEXT */

    extern OBJ ErrorNumber;

    if (_isString(aPathName)) {
        if (access(_stringVal(aPathName), W_OK) == 0) {
            RETURN ( true );
        }
        ErrorNumber = _MKSMALLINT(errno);
        RETURN ( false );
    }
%}
.
    self primitiveFailed
!

isExecutable:aPathName
    "return true, if the given file is executable"

%{  /* NOCONTEXT */

    extern OBJ ErrorNumber;

    if (_isString(aPathName)) {
        if (access(_stringVal(aPathName), X_OK) == 0) {
            RETURN ( true );
        }
        ErrorNumber = _MKSMALLINT(errno);
        RETURN ( false );
    }
%}
.
    self primitiveFailed
!

infoOf:aPathName
    "return an dictionary filled with info for the file 'aPathName';
     info is: (type->t mode->n uid->u gid->g size->s id->ino).
     return nil if such a file does not exist. A dictionary is returned,
     since we might need to add more info in the future without affecting
     existing applications."

    |info type mode uid gid size id|

    "{ Symbol: directory }"
    "{ Symbol: regular }"
    "{ Symbol: characterSpecial }"
    "{ Symbol: blockSpecial }"
    "{ Symbol: fifo }"
    "{ Symbol: socket }"
    "{ Symbol: symbolicLink }"
    "{ Symbol: unknown }"

%{
    struct stat buf;

    if (_isString(aPathName)) {
        if (stat((char *) _stringVal(aPathName), &buf) < 0) {
            RETURN ( nil );
        }
        switch (buf.st_mode & S_IFMT) {
            case S_IFDIR:
                type = _directory;
                break;
            case S_IFCHR:
                type = _characterSpecial;
                break;
            case S_IFBLK:
                type = _blockSpecial;
                break;
            case S_IFREG:
                type = _regular;
                break;
#ifdef S_IFLNK
            case S_IFLNK:
                type = _symbolicLink;
                break;
#endif
#ifdef S_IFSOCK
            case S_IFSOCK:
                type = _socket;
                break;
#endif
#ifdef S_IFIFO
            case S_IFIFO:
                type = _fifo;
                break;
#endif
            default:
                type = _unknown;
                break;
        }
        mode = _MKSMALLINT(buf.st_mode & 0777);
        uid = _MKSMALLINT(buf.st_uid);
        gid = _MKSMALLINT(buf.st_gid);
        size = _MKSMALLINT(buf.st_size);
        id = _MKSMALLINT(buf.st_ino);
    }
%}
.
    mode notNil ifTrue:[
        info := IdentityDictionary new.
        info at:#type put:type.
        info at:#mode put:mode.
        info at:#uid put:uid.
        info at:#gid put:gid.
        info at:#size put:size.
        info at:#id put:id.
        ^ info
   ].
   self primitiveFailed

   "OperatingSystem infoOf:'/'"
   "(OperatingSystem infoOf:'/') at:#uid"
!

accessModeOf:aPathName
    "return a number representing access rights rwxrwxrwx for owner,
     group and others. return nil if such a file does not exist."

    "
     this could have been implemented as:
        (self infoOf:aPathName) at:#mode
     but for huge directory searches the code below is faster
    "

%{  /* NOCONTEXT */

    struct stat buf;

    if (_isString(aPathName)) {
        if (stat((char *) _stringVal(aPathName), &buf) < 0) {
            RETURN ( nil );
        }
        RETURN ( _MKSMALLINT(buf.st_mode & 0777) );
    }
%}
.
   self primitiveFailed

   "(OperatingSystem accessModeOf:'/') printStringRadix:8"
!

changeAccessModeOf:aPathName to:modeBits
    "change the access rights rwxrwxrwx for owner,
     group and others of aPathName. return true if changed, false
     if such a file does not exist or change was not allowd."

%{  /* NOCONTEXT */

    if (_isString(aPathName) && _isSmallInteger(modeBits)) {
        RETURN ( (chmod((char *) _stringVal(aPathName), _intVal(modeBits) ) < 0) ?
                                false : true );
    }
%}
.
   self primitiveFailed
!

timeOfLastChange:aPathName
    "return the timeStamp of a file"

    |timeLow timeHi|
%{
    struct stat buf;
    time_t mtime;

    if (_isString(aPathName)) {
        if (stat((char *) _stringVal(aPathName), &buf) >= 0) {
            timeLow = _MKSMALLINT(buf.st_mtime & 0xFFFF);
            timeHi = _MKSMALLINT((buf.st_mtime >> 16) & 0xFFFF);
        }
    }
%}
.
    timeLow notNil ifTrue:[^ Time fromUnixTimeLow:timeLow and:timeHi].
    self primitiveFailed

    "OperatingSystem timeOfLastChange:'/'"
!

idOf:aPathName
    "return the fileNumber (i.e. inode number) of a file"

    "
     this could have been implemented as:
        (self infoOf:aPathName) at:#id 
     but for huge directory searches the code below is faster
    "

%{  /* NOCONTEXT */

    struct stat buf;

    if (_isString(aPathName)) {
        if (stat((char *) _stringVal(aPathName), &buf) >= 0) {
            RETURN (_MKSMALLINT(buf.st_ino));
        }
    }
%}
.
    self primitiveFailed

    "OperatingSystem idOf:'/'"
!

typeOf:aPathName
    "return the type of a file as a symbol"

    "
     this could have been implemented as:
        (self infoOf:aPathName) at:#type 
     but for huge directory searches the code below is faster
    "

%{  /* NOCONTEXT */

    struct stat buf;

    if (_isString(aPathName)) {
        if (stat((char *) _stringVal(aPathName), &buf) < 0) {
            RETURN ( nil );
        }
        switch (buf.st_mode & S_IFMT) {
            case S_IFDIR:
                RETURN ( _directory );
            case S_IFCHR:
                RETURN ( _characterSpecial );
            case S_IFBLK:
                RETURN ( _blockSpecial );
            case S_IFREG:
                RETURN ( _regular );
#ifdef S_IFLNK
            case S_IFLNK:
                RETURN ( _symbolicLink );
#endif
#ifdef S_IFSOCK
            case S_IFSOCK:
                RETURN ( _socket );
#endif
#ifdef S_IFIFO
            case S_IFIFO:
                RETURN ( _fifo );
#endif
            default:
                RETURN ( _unknown );
        }
    }
%}
.
    self primitiveFailed

    "OperatingSystem typeOf:'/'"
!

createDirectory:newPathName
    "create a new directory with name 'newPathName'.
     Return true if successful, false if failed."

    "since createDirectory is not used too often,
     you'll forgive me using mkdir ..."

    ^ self executeCommand:('mkdir ' , newPathName)

    "OperatingSystem createDirectory:'foo'"
!

recursiveCreateDirectory:dirName
    "create a directory - with all parent dirs if needed.
     Return true if successful, false otherwise. If false
     is returned, a partial created tree may be left,
     which is not cleaned-up here."

    self createDirectory:dirName.
    (self isDirectory:dirName) ifFalse:[
        (self recursiveCreateDirectory:(self directoryNameOf:dirName)) ifFalse:[^ false].
        ^ self createDirectory:dirName
    ].
    ^ (self isDirectory:dirName)

    "OperatingSystem recursiveCreateDirectory:'foo/bar/baz'"
!

removeFile:fullPathName
    "remove the file named 'fullPathName'; return true if successful"

%{  /* NOCONTEXT */

    if (_isString(fullPathName)) {
        RETURN ( (unlink((char *) _stringVal(fullPathName)) >= 0) ? true : false );
    }
%}
.
    self primitiveFailed
!

removeDirectory:fullPathName
    "remove the directory named 'fullPathName'.
     Return true if successful, false if directory is not empty or no permission"

%{  /* NOCONTEXT */

    if (_isString(fullPathName)) {
        RETURN ( (rmdir((char *) _stringVal(fullPathName)) >= 0) ? true : false );
    }
%}
.
    self primitiveFailed
!

link:oldPath to:newPath
    "link the file 'oldPath' to 'newPath'. The link will be a hard link.
     Return true if successful, false if not."

%{  /* NOCONTEXT */

    if (_isString(oldPath) && _isString(newPath)) {
        RETURN ( (link((char *) _stringVal(oldPath), (char *) _stringVal(newPath)) >= 0) ?
                                true : false );
    }
%}
.
    self primitiveFailed

    "OperatingSystem link:'foo' to:'bar'"
!

rename:oldPath to:newPath
    "rename the file 'oldPath' to 'newPath'. 
     Return true if sucessfull, false if not"

%{  /* NOCONTEXT */

    if (_isString(oldPath) && _isString(newPath)) {
#if defined(BSD)
        if (rename((char *) _stringVal(oldPath), (char *) _stringVal(newPath)) >= 0) {
            RETURN ( true );
        }
#else
        if (link((char *) _stringVal(oldPath), (char *) _stringVal(newPath)) >= 0) {
            if (unlink((char *) _stringVal(oldPath)) >= 0) {
                RETURN ( true );
            }
            unlink((char *) _stringVal(newPath));
        }
#endif
        RETURN ( false );
    }
%}
.
    self primitiveFailed

    "OperatingSystem rename:'foo' to:'bar'"
! !
